# cython: language_level=3
# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Package metadata."""

from __future__ import annotations

import typing

# DO NOT CHANGE THE TYPE HINTS FOR THESE FIELDS. THESE ARE AUTOMATICALLY UPDATED
# FROM THE CI SCRIPT AND DOING THIS MAY LEAD TO THE RELEASE PROCESS FAILING.

__author__: typing.Final[str] = "Nekokatt"
__maintainer__: typing.Final[str] = "davfsa"
__ci__: typing.Final[str] = "https://github.com/hikari-py/hikari/actions"
__copyright__: typing.Final[str] = "2021-present, davfsa"
__coverage__: typing.Final[str] = "https://codeclimate.com/github/hikari-py/hikari"
__discord_invite__: typing.Final[str] = "https://discord.gg/Jx4cNGG"
__docs__: typing.Final[str] = "https://docs.hikari-py.dev/en/2.1.1"
__email__: typing.Final[str] = "davfsa@gmail.com"
__issue_tracker__: typing.Final[str] = "https://github.com/hikari-py/hikari/issues"
__license__: typing.Final[str] = "MIT"
__url__: typing.Final[str] = "https://github.com/hikari-py/hikari"
__version__: typing.Final[str] = "2.1.1"
__git_sha1__: typing.Final[str] = "4200bc4b17832d8f2194c5735f4f387bf96ce359"
